
from typing import Tuple, Generator
import sys
from itertools import product
from functools import reduce

import numpy as np

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

from .utils import DuplicatedKeyError, WrongArrayDimensionException, WrongArrayShapeException


class NumpyNDArrayWrappedDict(dict):
    def __init__(
            self,
            lists_keystrings: list[list[str]],
            default_initial_value: float=0.0
    ):
        super(dict, self).__init__()
        for list_keystrings in lists_keystrings:
            if (len(list_keystrings)) != len(set(list_keystrings)):
                raise DuplicatedKeyError()
        self._lists_keystrings = lists_keystrings
        self._keystrings_to_indices = [
            {
                keyword: idx for idx, keyword in enumerate(list_keystrings)
            }
            for list_keystrings in self._lists_keystrings
        ]

        self._tensor_dimensions = len(self._lists_keystrings)
        self._dimension_sizes = [len(l) for l in self._lists_keystrings]
        self._total_size = reduce(lambda a, b: a*b, self._dimension_sizes)

        self._numpyarray = np.empty(tuple(len(l) for l in self._lists_keystrings))
        self._numpyarray.fill(default_initial_value)

    def _get_indices(self, item: Tuple[str, ...]) -> list[int]:
        return [
            mapping[keyword]
            for mapping, keyword in zip(self._keystrings_to_indices, item)
        ]

    def __getitem__(self, item: Tuple[str, ...]) -> float:
        if len(item) != self.tensor_dimensions:
            raise WrongArrayDimensionException(self.tensor_dimensions, len(item))
        indices = self._get_indices(item)
        return self._numpyarray[tuple(indices)]

    def __setitem__(self, key: Tuple[str, ...], value: float) -> None:
        if len(key) != self.tensor_dimensions:
            raise WrongArrayDimensionException(self.tensor_dimensions, len(key))
        indices = self._get_indices(key)
        self._numpyarray[tuple(indices)] = value

    def update(self, new_dict: dict):
        raise TypeError("We cannot update this kind of dict this way!")

    def __iter__(self) -> Generator[Tuple[str, ...], None, None]:
        for keywords_tuple in product(*self._lists_keystrings):
            yield keywords_tuple

    def keys(self):
        return list(self.__iter__())

    def values(self):
        return [self.__getitem__(keywords_tuple) for keywords_tuple in self.__iter__()]

    def items(self):
        return [
            (keywords_tuple, self.__getitem__(keywords_tuple))
            for keywords_tuple in self.__iter__()
        ]

    def to_numpy(self) -> np.ndarray:
        return self._numpyarray

    def generate_dict(self, nparray: np.ndarray) -> Self:
        if len(nparray.shape) != self.tensor_dimensions:
            raise WrongArrayDimensionException(self.tensor_dimensions, len(nparray.shape))
        if nparray.shape != self._numpyarray.shape:
            raise WrongArrayShapeException(self._numpyarray.shape, nparray.shape)
        wrapped_dict = NumpyNDArrayWrappedDict(self._lists_keystrings)
        wrapped_dict._numpyarray = nparray
        return wrapped_dict

    def __repr__(self) -> str:
        return f"<NumpyNDArrayWrappedDict: dimensions ({', '.join(map(str, self.dimension_sizes))})>"

    def __str__(self) -> str:
        return self.__repr__()

    def __len__(self) -> int:
        return self._total_size

    def to_dict(self) -> dict[Tuple[str, ...], float]:
        return {
            keywords_tuple: value for keywords_tuple, value in self.items()
        }

    @classmethod
    def from_dict_given_keywords(
            cls,
            lists_keywords: list[list[str]],
            oridict: dict[Tuple[str, ...], float],
            default_initial_value: float = 0.0
    ) -> Self:
        wrapped_dict = NumpyNDArrayWrappedDict(
            lists_keywords,
            default_initial_value=default_initial_value
        )
        for keywords_tuple in product(*lists_keywords):
            wrapped_dict[keywords_tuple] = oridict.get(keywords_tuple, default_initial_value)
        return wrapped_dict

    @classmethod
    def from_dict(
            cls,
            oridict: dict[Tuple[str, ...], float],
            default_initial_value: float = 0.0
    ) -> Self:
        nbdims = len(next(iter(oridict)))
        lists_keystrings = [
            list(set(keystring[i] for keystring in oridict.keys()))
            for i in range(nbdims)
        ]
        return cls.from_dict_given_keywords(
            lists_keystrings,
            oridict,
            default_initial_value=default_initial_value
        )

    @property
    def tensor_dimensions(self) -> int:
        return self._tensor_dimensions

    @property
    def dimension_sizes(self) -> list[int]:
        return self._dimension_sizes





from typing import Tuple


class DuplicatedKeyError(Exception):
    def __init__(self):
        self.message = "Duplicated keys!"


class WrongArrayDimensionException(Exception):
    def __init__(self, expected_dimension: int, given_dimensions: int):
        self.message = f"Expected dimension: {expected_dimension}, but {given_dimensions} dimensions are given!"


class WrongArrayShapeException(Exception):
    def __init__(self, expected_shape: Tuple[int, ...], given_shape: Tuple[int, ...]):
        self.message = f"Expected shape: {', '.join(str(dim_len) for dim_len in expected_shape)}, but the given array shape is {', '.join(str(dim_len) for dim_len in given_shape)}!"




from functools import reduce
from typing import Tuple, Union
import sys
from itertools import product

import numpy as np
import sparse

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

from .wrap import NumpyNDArrayWrappedDict
from .utils import DuplicatedKeyError, WrongArrayDimensionException, WrongArrayShapeException


class SparseArrayWrappedDict(NumpyNDArrayWrappedDict):
    def __init__(
            self,
            lists_keystrings: list[list[str]],
            default_initial_value: float=0.0
    ):
        super(dict, self).__init__()
        for list_keystrings in lists_keystrings:
            if (len(list_keystrings)) != len(set(list_keystrings)):
                raise DuplicatedKeyError()
        self._lists_keystrings = lists_keystrings
        self._keystrings_to_indices = [
            {
                keyword: idx for idx, keyword in enumerate(list_keystrings)
            }
            for list_keystrings in self._lists_keystrings
        ]

        self._tensor_dimensions = len(self._lists_keystrings)
        self._dimension_sizes = [len(l) for l in self._lists_keystrings]
        self._total_size = reduce(lambda a, b: a*b, self._dimension_sizes)

        self._sparsearray = sparse.DOK(
            tuple(len(l) for l in self._lists_keystrings),
            fill_value=default_initial_value
        )

    def __getitem__(self, item: Tuple[str, ...]) -> float:
        if len(item) != self.tensor_dimensions:
            raise WrongArrayDimensionException(self.tensor_dimensions, len(item))
        indices = self._get_indices(item)
        return self._sparsearray[tuple(indices)]

    def __setitem__(self, key: Tuple[str, ...], value: float) -> None:
        if len(key) != self.tensor_dimensions:
            raise WrongArrayDimensionException(self.tensor_dimensions, len(key))
        indices = self._get_indices(key)
        self._sparsearray[tuple(indices)] = value

    def to_numpy(self) -> np.ndarray:
        return self._sparsearray.todense()

    def to_coo(self) -> sparse.COO:
        return self._sparsearray.to_coo()

    def to_dok(self) -> sparse.DOK:
        return self._sparsearray

    def generate_dict(
            self,
            new_array: Union[np.ndarray, sparse.SparseArray],
            dense: bool=False
    ) -> Self:
        if len(new_array.shape) != self.tensor_dimensions:
            raise WrongArrayDimensionException(self.tensor_dimensions, len(new_array.shape))
        if new_array.shape != self._sparsearray.shape:
            raise WrongArrayShapeException(self._sparsearray.shape, new_array.shape)
        if dense:
            wrapped_dict = NumpyNDArrayWrappedDict(self._lists_keystrings)
            if isinstance(new_array, sparse.SparseArray):
                wrapped_dict._numpyarray = new_array.todense()
            else:
                wrapped_dict._numpyarray = new_array
        else:
            wrapped_dict = SparseArrayWrappedDict(self._lists_keystrings)
            if isinstance(new_array, sparse.SparseArray):
                wrapped_dict._sparsearray = new_array if isinstance(new_array, sparse.DOK) else sparse.DOK(new_array)
            else:
                wrapped_dict._sparsearray = sparse.DOK(new_array)
        return wrapped_dict

    def __repr__(self) -> str:
        return f"<SparseArrayWrappedDict: dimensions ({', '.join(map(str, self.dimension_sizes))})>"

    @classmethod
    def from_dict_given_keywords(
            cls,
            lists_keywords: list[list[str]],
            oridict: dict[Tuple[str, ...], float],
            default_initial_value: float = 0.0
    ) -> Self:
        wrapped_dict = SparseArrayWrappedDict(
            lists_keywords,
            default_initial_value=default_initial_value
        )
        for keywords_tuple in product(*lists_keywords):
            wrapped_dict[keywords_tuple] = oridict.get(keywords_tuple, default_initial_value)
        return wrapped_dict




import unittest
import numpy as np
import sparse

from npdict import SparseArrayWrappedDict
from npdict.utils import DuplicatedKeyError, WrongArrayDimensionException, WrongArrayShapeException


class TestSparseArrayWrappedDict(unittest.TestCase):

    def setUp(self):
        self.lists_keystrings = [
            ['a', 'b', 'c'],
            ['d', 'e']
        ]
        self.wrapped_dict = SparseArrayWrappedDict(self.lists_keystrings, default_initial_value=1.0)

    def test_initialization(self):
        self.assertEqual(self.wrapped_dict.tensor_dimensions, 2)
        self.assertEqual(self.wrapped_dict.dimension_sizes, [3, 2])
        self.assertEqual(len(self.wrapped_dict), 6)
        np.testing.assert_array_equal(self.wrapped_dict.to_numpy(), np.ones((3, 2)))

    def test_initialization_with_duplicated_keys(self):
        with self.assertRaises(DuplicatedKeyError):
            SparseArrayWrappedDict([['a', 'a'], ['b', 'c']])

    def test_getitem_setitem(self):
        self.wrapped_dict[('a', 'd')] = 2.0
        self.assertEqual(self.wrapped_dict[('a', 'd')], 2.0)
        self.assertEqual(self.wrapped_dict[('b', 'e')], 1.0)

    def test_getitem_wrong_dimension(self):
        with self.assertRaises(WrongArrayDimensionException):
            _ = self.wrapped_dict[('a',)]

    def test_setitem_wrong_dimension(self):
        with self.assertRaises(WrongArrayDimensionException):
            self.wrapped_dict[('a', 'b', 'c')] = 1.0

    def test_to_numpy(self):
        self.wrapped_dict[('a', 'd')] = 3.0
        expected_array = np.array([[3., 1.], [1., 1.], [1., 1.]])
        np.testing.assert_array_equal(self.wrapped_dict.to_numpy(), expected_array)

    def test_to_coo(self):
        self.wrapped_dict[('b', 'e')] = 4.0
        coo = self.wrapped_dict.to_coo()
        self.assertIsInstance(coo, sparse.COO)
        self.assertEqual(coo[1, 1], 4.0)
        self.assertEqual(coo[0, 1], 1.0)

    def test_to_dok(self):
        self.wrapped_dict[('c', 'd')] = 5.0
        dok = self.wrapped_dict.to_dok()
        self.assertIsInstance(dok, sparse.DOK)
        self.assertEqual(dok[2, 0], 5.0)
        self.assertEqual(dok[1, 1], 1.0)

    def test_generate_dict(self):
        new_sparse_array = sparse.random((3, 2), density=0.5)
        
        # Test generating a new sparse dict
        new_sparse_dict = self.wrapped_dict.generate_dict(new_sparse_array, dense=False)
        self.assertIsInstance(new_sparse_dict, SparseArrayWrappedDict)
        np.testing.assert_array_equal(new_sparse_dict.to_numpy(), new_sparse_array.todense())

        # Test generating a new dense dict
        new_dense_dict = self.wrapped_dict.generate_dict(new_sparse_array, dense=True)
        from npdict import NumpyNDArrayWrappedDict
        self.assertIsInstance(new_dense_dict, NumpyNDArrayWrappedDict)
        np.testing.assert_array_equal(new_dense_dict.to_numpy(), new_sparse_array.todense())

    def test_generate_dict_wrong_shape(self):
        new_array = sparse.DOK((2, 3))
        with self.assertRaises(WrongArrayShapeException):
            self.wrapped_dict.generate_dict(new_array)

    def test_repr(self):
        self.assertEqual(repr(self.wrapped_dict), f"<SparseArrayWrappedDict: dimensions ({', '.join(map(str, self.wrapped_dict.dimension_sizes))})>")

    def test_from_dict_given_keywords(self):
        d = {('a', 'd'): 10.0, ('c', 'e'): 20.0}
        keywords = [['a', 'b', 'c'], ['d', 'e']]
        wrapped = SparseArrayWrappedDict.from_dict_given_keywords(keywords, d, default_initial_value=-1.0)
        self.assertEqual(wrapped[('a', 'd')], 10.0)
        self.assertEqual(wrapped[('c', 'e')], 20.0)
        self.assertEqual(wrapped[('b', 'd')], -1.0)


if __name__ == '__main__':
    unittest.main()


import unittest
import numpy as np

from npdict import NumpyNDArrayWrappedDict
from npdict.utils import DuplicatedKeyError, WrongArrayDimensionException, WrongArrayShapeException


class TestNumpyNDArrayWrappedDict(unittest.TestCase):

    def setUp(self):
        self.lists_keystrings = [
            ['a', 'b', 'c'],
            ['d', 'e']
        ]
        self.wrapped_dict = NumpyNDArrayWrappedDict(self.lists_keystrings, default_initial_value=1.0)

    def test_initialization(self):
        self.assertEqual(self.wrapped_dict.tensor_dimensions, 2)
        self.assertEqual(self.wrapped_dict.dimension_sizes, [3, 2])
        self.assertEqual(len(self.wrapped_dict), 6)
        np.testing.assert_array_equal(self.wrapped_dict.to_numpy(), np.ones((3, 2)))

    def test_initialization_with_duplicated_keys(self):
        with self.assertRaises(DuplicatedKeyError):
            NumpyNDArrayWrappedDict([['a', 'a'], ['b', 'c']])

    def test_getitem(self):
        self.wrapped_dict[('a', 'd')] = 2.0
        self.assertEqual(self.wrapped_dict[('a', 'd')], 2.0)
        self.wrapped_dict[('a', 'e')] = 1.0   # test empty item

    def test_getitem_wrong_dimension(self):
        with self.assertRaises(WrongArrayDimensionException):
            _ = self.wrapped_dict[('a',)]

    def test_setitem(self):
        self.wrapped_dict[('b', 'e')] = 3.0
        self.assertEqual(self.wrapped_dict[('b', 'e')], 3.0)
        self.assertEqual(self.wrapped_dict.to_numpy()[1, 1], 3.0)

    def test_setitem_wrong_dimension(self):
        with self.assertRaises(WrongArrayDimensionException):
            self.wrapped_dict[('a', 'b', 'c')] = 1.0

    def test_iteration(self):
        keys = list(self.wrapped_dict)
        self.assertEqual(len(keys), 6)
        self.assertIn(('a', 'd'), keys)
        self.assertIn(('c', 'e'), keys)

    def test_keys(self):
        keys = self.wrapped_dict.keys()
        self.assertEqual(len(keys), 6)
        self.assertIn(('a', 'd'), keys)

    def test_values(self):
        self.wrapped_dict[('a', 'd')] = 5.0
        self.wrapped_dict[('c', 'e')] = 6.0
        values = self.wrapped_dict.values()
        self.assertIn(1.0, values)
        self.assertIn(5.0, values)
        self.assertIn(6.0, values)

    def test_items(self):
        self.wrapped_dict[('a', 'd')] = 7.0
        items = self.wrapped_dict.items()
        self.assertIn((('a', 'd'), 7.0), items)
        self.assertIn((('b', 'd'), 1.0), items)

    def test_to_dict(self):
        self.wrapped_dict[('a', 'd')] = 8.0
        d = self.wrapped_dict.to_dict()
        self.assertEqual(d[('a', 'd')], 8.0)
        self.assertEqual(d[('b', 'e')], 1.0)

    def test_from_dict(self):
        d = {('a', 'x'): 1, ('b', 'y'): 2}
        wrapped = NumpyNDArrayWrappedDict.from_dict(d)
        self.assertEqual(wrapped[('a', 'x')], 1)
        self.assertEqual(wrapped[('b', 'y')], 2)
        self.assertEqual(wrapped[('a', 'y')], 0.0)
        self.assertEqual(wrapped[('b', 'x')], 0.0)

    def test_from_dict_given_keywords(self):
        d = {('a', 'x'): 1}
        keywords = [['a', 'b'], ['x', 'y']]
        wrapped = NumpyNDArrayWrappedDict.from_dict_given_keywords(keywords, d, default_initial_value=-1)
        self.assertEqual(wrapped[('a', 'x')], 1)
        self.assertEqual(wrapped[('b', 'y')], -1)

    def test_repr_str(self):
        self.assertEqual(repr(self.wrapped_dict), f"<NumpyNDArrayWrappedDict: dimensions ({', '.join(map(str, self.wrapped_dict.dimension_sizes))})>")
        self.assertEqual(str(self.wrapped_dict), repr(self.wrapped_dict))

    def test_update(self):
        with self.assertRaises(TypeError):
            self.wrapped_dict.update({('a', 'd'): 1})

    def test_generate_dict(self):
        new_array = np.zeros((3, 2))
        new_wrapped_dict = self.wrapped_dict.generate_dict(new_array)
        np.testing.assert_array_equal(new_wrapped_dict.to_numpy(), new_array)
        self.assertEqual(new_wrapped_dict.dimension_sizes, self.wrapped_dict.dimension_sizes)

    def test_generate_dict_wrong_shape(self):
        new_array = np.zeros((2, 3))
        with self.assertRaises(WrongArrayShapeException):
            self.wrapped_dict.generate_dict(new_array)


if __name__ == '__main__':
    unittest.main()
